# Copyright (C) 2015-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import abc
import functools
import inspect
import logging
import pickle
import requests

from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    Union,
)

from flask import Flask, Request, Response, request, abort
from werkzeug.exceptions import HTTPException

from .serializers import (
    decode_response,
    encode_data_client as encode_data,
    msgpack_dumps,
    msgpack_loads,
    json_dumps,
    json_loads,
    exception_to_dict,
)

from .negotiation import (
    Formatter as FormatterBase,
    Negotiator as NegotiatorBase,
    negotiate as _negotiate,
)


logger = logging.getLogger(__name__)


# support for content negotiation


class Negotiator(NegotiatorBase):
    def best_mimetype(self):
        return request.accept_mimetypes.best_match(
            self.accept_mimetypes, "application/json"
        )

    def _abort(self, status_code, err=None):
        return abort(status_code, err)


def negotiate(formatter_cls, *args, **kwargs):
    return _negotiate(Negotiator, formatter_cls, *args, **kwargs)


class Formatter(FormatterBase):
    def _make_response(self, body, content_type):
        return Response(body, content_type=content_type)

    def configure(self, extra_encoders=None):
        self.extra_encoders = extra_encoders


class JSONFormatter(Formatter):
    format = "json"
    mimetypes = ["application/json"]

    def render(self, obj):
        return json_dumps(obj, extra_encoders=self.extra_encoders)


class MsgpackFormatter(Formatter):
    format = "msgpack"
    mimetypes = ["application/x-msgpack"]

    def render(self, obj):
        return msgpack_dumps(obj, extra_encoders=self.extra_encoders)


# base API classes


class RemoteException(Exception):
    """raised when remote returned an out-of-band failure notification, e.g., as a
    HTTP status code or serialized exception

    Attributes:
        response: HTTP response corresponding to the failure

    """

    def __init__(
        self,
        payload: Optional[Any] = None,
        response: Optional[requests.Response] = None,
    ):
        if payload is not None:
            super().__init__(payload)
        else:
            super().__init__()
        self.response = response

    def __str__(self):
        if (
            self.args
            and isinstance(self.args[0], dict)
            and "type" in self.args[0]
            and "args" in self.args[0]
        ):
            return (
                f"<RemoteException {self.response.status_code} "
                f'{self.args[0]["type"]}: {self.args[0]["args"]}>'
            )
        else:
            return super().__str__()


def remote_api_endpoint(path):
    def dec(f):
        f._endpoint_path = path
        return f

    return dec


class APIError(Exception):
    """API Error"""

    def __str__(self):
        return "An unexpected error occurred in the backend: {}".format(self.args)


class MetaRPCClient(type):
    """Metaclass for RPCClient, which adds a method for each endpoint
    of the database it is designed to access.

    See for example :class:`swh.indexer.storage.api.client.RemoteStorage`"""

    def __new__(cls, name, bases, attributes):
        # For each method wrapped with @remote_api_endpoint in an API backend
        # (eg. :class:`swh.indexer.storage.IndexerStorage`), add a new
        # method in RemoteStorage, with the same documentation.
        #
        # Note that, despite the usage of decorator magic (eg. functools.wrap),
        # this never actually calls an IndexerStorage method.
        backend_class = attributes.get("backend_class", None)
        for base in bases:
            if backend_class is not None:
                break
            backend_class = getattr(base, "backend_class", None)
        if backend_class:
            for (meth_name, meth) in backend_class.__dict__.items():
                if hasattr(meth, "_endpoint_path"):
                    cls.__add_endpoint(meth_name, meth, attributes)
        return super().__new__(cls, name, bases, attributes)

    @staticmethod
    def __add_endpoint(meth_name, meth, attributes):
        wrapped_meth = inspect.unwrap(meth)

        @functools.wraps(meth)  # Copy signature and doc
        def meth_(*args, **kwargs):
            # Match arguments and parameters
            post_data = inspect.getcallargs(wrapped_meth, *args, **kwargs)

            # Remove arguments that should not be passed
            self = post_data.pop("self")
            post_data.pop("cur", None)
            post_data.pop("db", None)

            # Send the request.
            return self.post(meth._endpoint_path, post_data)

        if meth_name not in attributes:
            attributes[meth_name] = meth_


class RPCClient(metaclass=MetaRPCClient):
    """Proxy to an internal SWH RPC

    """

    backend_class = None  # type: ClassVar[Optional[type]]
    """For each method of `backend_class` decorated with
    :func:`remote_api_endpoint`, a method with the same prototype and
    docstring will be added to this class. Calls to this new method will
    be translated into HTTP requests to a remote server.

    This backend class will never be instantiated, it only serves as
    a template."""

    api_exception = APIError  # type: ClassVar[Type[Exception]]
    """The exception class to raise in case of communication error with
    the server."""

    reraise_exceptions: ClassVar[List[Type[Exception]]] = []
    """On server errors, if any of the exception classes in this list
    has the same name as the error name, then the exception will
    be instantiated and raised instead of a generic RemoteException."""

    extra_type_encoders: List[Tuple[type, str, Callable]] = []
    """Value of `extra_encoders` passed to `json_dumps` or `msgpack_dumps`
    to be able to serialize more object types."""
    extra_type_decoders: Dict[str, Callable] = {}
    """Value of `extra_decoders` passed to `json_loads` or `msgpack_loads`
    to be able to deserialize more object types."""

    def __init__(
        self,
        url,
        api_exception=None,
        timeout=None,
        chunk_size=4096,
        reraise_exceptions=None,
        **kwargs,
    ):
        if api_exception:
            self.api_exception = api_exception
        if reraise_exceptions:
            self.reraise_exceptions = reraise_exceptions
        base_url = url if url.endswith("/") else url + "/"
        self.url = base_url
        self.session = requests.Session()
        adapter = requests.adapters.HTTPAdapter(
            max_retries=kwargs.get("max_retries", 3),
            pool_connections=kwargs.get("pool_connections", 20),
            pool_maxsize=kwargs.get("pool_maxsize", 100),
        )
        self.session.mount(self.url, adapter)

        self.timeout = timeout
        self.chunk_size = chunk_size

    def _url(self, endpoint):
        return "%s%s" % (self.url, endpoint)

    def raw_verb(self, verb, endpoint, **opts):
        if "chunk_size" in opts:
            # if the chunk_size argument has been passed, consider the user
            # also wants stream=True, otherwise, what's the point.
            opts["stream"] = True
        if self.timeout and "timeout" not in opts:
            opts["timeout"] = self.timeout
        try:
            return getattr(self.session, verb)(self._url(endpoint), **opts)
        except requests.exceptions.ConnectionError as e:
            raise self.api_exception(e)

    def post(self, endpoint, data, **opts):
        if isinstance(data, (abc.Iterator, abc.Generator)):
            data = (self._encode_data(x) for x in data)
        else:
            data = self._encode_data(data)
        chunk_size = opts.pop("chunk_size", self.chunk_size)
        response = self.raw_verb(
            "post",
            endpoint,
            data=data,
            headers={
                "content-type": "application/x-msgpack",
                "accept": "application/x-msgpack",
            },
            **opts,
        )
        if opts.get("stream") or response.headers.get("transfer-encoding") == "chunked":
            self.raise_for_status(response)
            return response.iter_content(chunk_size)
        else:
            return self._decode_response(response)

    def _encode_data(self, data):
        return encode_data(data, extra_encoders=self.extra_type_encoders)

    post_stream = post

    def get(self, endpoint, **opts):
        chunk_size = opts.pop("chunk_size", self.chunk_size)
        response = self.raw_verb(
            "get", endpoint, headers={"accept": "application/x-msgpack"}, **opts
        )
        if opts.get("stream") or response.headers.get("transfer-encoding") == "chunked":
            self.raise_for_status(response)
            return response.iter_content(chunk_size)
        else:
            return self._decode_response(response)

    def get_stream(self, endpoint, **opts):
        return self.get(endpoint, stream=True, **opts)

    def raise_for_status(self, response) -> None:
        """check response HTTP status code and raise an exception if it denotes an
        error; do nothing otherwise

        """
        status_code = response.status_code
        status_class = response.status_code // 100

        if status_code == 404:
            raise RemoteException(payload="404 not found", response=response)

        exception = None

        # TODO: only old servers send pickled error; stop trying to unpickle
        # after they are all upgraded
        try:
            if status_class == 4:
                data = self._decode_response(response, check_status=False)
                if isinstance(data, dict):
                    for exc_type in self.reraise_exceptions:
                        if exc_type.__name__ == data["exception"]["type"]:
                            exception = exc_type(*data["exception"]["args"])
                            break
                    else:
                        exception = RemoteException(
                            payload=data["exception"], response=response
                        )
                else:
                    exception = pickle.loads(data)

            elif status_class == 5:
                data = self._decode_response(response, check_status=False)
                if "exception_pickled" in data:
                    exception = pickle.loads(data["exception_pickled"])
                else:
                    exception = RemoteException(
                        payload=data["exception"], response=response
                    )

        except (TypeError, pickle.UnpicklingError):
            raise RemoteException(payload=data, response=response)

        if exception:
            raise exception from None

        if status_class != 2:
            raise RemoteException(
                payload=f"API HTTP error: {status_code} {response.content}",
                response=response,
            )

    def _decode_response(self, response, check_status=True):
        if check_status:
            self.raise_for_status(response)
        return decode_response(response, extra_decoders=self.extra_type_decoders)

    def __repr__(self):
        return "<{} url={}>".format(self.__class__.__name__, self.url)


class BytesRequest(Request):
    """Request with proper escaping of arbitrary byte sequences."""

    encoding = "utf-8"
    encoding_errors = "surrogateescape"


ENCODERS: Dict[str, Callable[[Any], Union[bytes, str]]] = {
    "application/x-msgpack": msgpack_dumps,
    "application/json": json_dumps,
}


def encode_data_server(
    data, content_type="application/x-msgpack", extra_type_encoders=None
):
    encoded_data = ENCODERS[content_type](data, extra_encoders=extra_type_encoders)
    return Response(encoded_data, mimetype=content_type,)


def decode_request(request, extra_decoders=None):
    content_type = request.mimetype
    data = request.get_data()
    if not data:
        return {}

    if content_type == "application/x-msgpack":
        r = msgpack_loads(data, extra_decoders=extra_decoders)
    elif content_type == "application/json":
        # XXX this .decode() is needed for py35.
        # Should not be needed any more with py37
        r = json_loads(data.decode("utf-8"), extra_decoders=extra_decoders)
    else:
        raise ValueError("Wrong content type `%s` for API request" % content_type)

    return r


def error_handler(exception, encoder, status_code=500):
    logging.exception(exception)
    response = encoder(exception_to_dict(exception))
    if isinstance(exception, HTTPException):
        response.status_code = exception.code
    else:
        # TODO: differentiate between server errors and client errors
        response.status_code = status_code
    return response


class RPCServerApp(Flask):
    """For each endpoint of the given `backend_class`, tells app.route to call
    a function that decodes the request and sends it to the backend object
    provided by the factory.

    :param Any backend_class:
        The class of the backend, which will be analyzed to look
        for API endpoints.
    :param Optional[Callable[[], backend_class]] backend_factory:
        A function with no argument that returns an instance of
        `backend_class`. If unset, defaults to calling `backend_class`
        constructor directly.
    """

    request_class = BytesRequest

    extra_type_encoders: List[Tuple[type, str, Callable]] = []
    """Value of `extra_encoders` passed to `json_dumps` or `msgpack_dumps`
    to be able to serialize more object types."""
    extra_type_decoders: Dict[str, Callable] = {}
    """Value of `extra_decoders` passed to `json_loads` or `msgpack_loads`
    to be able to deserialize more object types."""

    def __init__(self, *args, backend_class=None, backend_factory=None, **kwargs):
        super().__init__(*args, **kwargs)

        self.backend_class = backend_class
        if backend_class is not None:
            if backend_factory is None:
                backend_factory = backend_class
            for (meth_name, meth) in backend_class.__dict__.items():
                if hasattr(meth, "_endpoint_path"):
                    self.__add_endpoint(meth_name, meth, backend_factory)

    def __add_endpoint(self, meth_name, meth, backend_factory):
        from flask import request

        @self.route("/" + meth._endpoint_path, methods=["POST"])
        @negotiate(MsgpackFormatter, extra_encoders=self.extra_type_encoders)
        @negotiate(JSONFormatter, extra_encoders=self.extra_type_encoders)
        @functools.wraps(meth)  # Copy signature and doc
        def _f():
            # Call the actual code
            obj_meth = getattr(backend_factory(), meth_name)
            kw = decode_request(request, extra_decoders=self.extra_type_decoders)
            return obj_meth(**kw)



# Copyright (C) 2016-2017  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import itertools
import codecs
import re

from contextlib import contextmanager


@contextmanager
def cwd(path):
    """Contextually change the working directory to do thy bidding.
    Then gets back to the original location.

    """
    prev_cwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(prev_cwd)


def grouper(iterable, n):
    """Collect data into fixed-length size iterables. The last block might
       contain less elements as it will hold only the remaining number
       of elements.

       The invariant here is that the number of elements in the input
       iterable and the sum of the number of elements of all iterables
       generated from this function should be equal.

    Args:
        iterable (Iterable): an iterable
        n (int): size of block to slice the iterable into

    Yields:
        fixed-length blocks as iterables. As mentioned, the last
        iterable might be less populated.

    """
    args = [iter(iterable)] * n
    stop_value = object()
    for _data in itertools.zip_longest(*args, fillvalue=stop_value):
        yield (d for d in _data if d is not stop_value)


def backslashescape_errors(exception):
    if isinstance(exception, UnicodeDecodeError):
        bad_data = exception.object[exception.start : exception.end]
        escaped = "".join(r"\x%02x" % x for x in bad_data)
        return escaped, exception.end

    return codecs.backslashreplace_errors(exception)


codecs.register_error("backslashescape", backslashescape_errors)


def encode_with_unescape(value):
    """Encode an unicode string containing \\x<hex> backslash escapes"""
    slices = []
    start = 0
    odd_backslashes = False
    i = 0
    while i < len(value):
        if value[i] == "\\":
            odd_backslashes = not odd_backslashes
        else:
            if odd_backslashes:
                if value[i] != "x":
                    raise ValueError(
                        "invalid escape for %r at position %d" % (value, i - 1)
                    )
                slices.append(
                    value[start : i - 1].replace("\\\\", "\\").encode("utf-8")
                )
                slices.append(bytes.fromhex(value[i + 1 : i + 3]))

                odd_backslashes = False
                start = i = i + 3
                continue

        i += 1

    slices.append(value[start:i].replace("\\\\", "\\").encode("utf-8"))

    return b"".join(slices)


def decode_with_escape(value):
    """Decode a bytestring as utf-8, escaping the bytes of invalid utf-8 sequences
    as \\x<hex value>. We also escape NUL bytes as they are invalid in JSON
    strings.
    """
    # escape backslashes
    value = value.replace(b"\\", b"\\\\")
    value = value.replace(b"\x00", b"\\x00")
    return value.decode("utf-8", "backslashescape")


def commonname(path0, path1, as_str=False):
    """Compute the commonname between the path0 and path1.

    """
    return path1.split(path0)[1]


def numfile_sortkey(fname):
    """Simple function to sort filenames of the form:

      nnxxx.ext

    where nn is a number according to the numbers.

    Typically used to sort sql/nn-swh-xxx.sql files.
    """
    num, rem = re.match(r"(\d*)(.*)", fname).groups()
    return (num and int(num) or 99, rem)



# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.core.api import remote_api_endpoint, RPCServerApp, RPCClient
from swh.core.api import error_handler, encode_data_server, RemoteException


# this class is used on the server part
class RPCTest:
    @remote_api_endpoint("endpoint_url")
    def endpoint(self, test_data, db=None, cur=None):
        assert test_data == "spam"
        return "egg"

    @remote_api_endpoint("path/to/endpoint")
    def something(self, data, db=None, cur=None):
        return data

    @remote_api_endpoint("raises_typeerror")
    def raise_typeerror(self):
        raise TypeError("Did I pass through?")


# this class is used on the client part. We cannot inherit from RPCTest
# because the automagic metaclass based code that generates the RPCClient
# proxy class from this does not handle inheritance properly.
# We do add an endpoint on the client side that has no implementation
# server-side to test this very situation (in should generate a 404)
class RPCTest2:
    @remote_api_endpoint("endpoint_url")
    def endpoint(self, test_data, db=None, cur=None):
        assert test_data == "spam"
        return "egg"

    @remote_api_endpoint("path/to/endpoint")
    def something(self, data, db=None, cur=None):
        return data

    @remote_api_endpoint("not_on_server")
    def not_on_server(self, db=None, cur=None):
        return "ok"

    @remote_api_endpoint("raises_typeerror")
    def raise_typeerror(self):
        return "data"


class RPCTestClient(RPCClient):
    backend_class = RPCTest2


@pytest.fixture
def app():
    # This fixture is used by the 'swh_rpc_adapter' fixture
    # which is defined in swh/core/pytest_plugin.py
    application = RPCServerApp("testapp", backend_class=RPCTest)

    @application.errorhandler(Exception)
    def my_error_handler(exception):
        return error_handler(exception, encode_data_server)

    return application


@pytest.fixture
def swh_rpc_client_class():
    # This fixture is used by the 'swh_rpc_client' fixture
    # which is defined in swh/core/pytest_plugin.py
    return RPCTestClient


def test_api_client_endpoint_missing(swh_rpc_client):
    with pytest.raises(AttributeError):
        swh_rpc_client.missing(data="whatever")


def test_api_server_endpoint_missing(swh_rpc_client):
    # A 'missing' endpoint (server-side) should raise an exception
    # due to a 404, since at the end, we do a GET/POST an inexistent URL
    with pytest.raises(Exception, match="404 not found"):
        swh_rpc_client.not_on_server()


def test_api_endpoint_kwargs(swh_rpc_client):
    res = swh_rpc_client.something(data="whatever")
    assert res == "whatever"
    res = swh_rpc_client.endpoint(test_data="spam")
    assert res == "egg"


def test_api_endpoint_args(swh_rpc_client):
    res = swh_rpc_client.something("whatever")
    assert res == "whatever"
    res = swh_rpc_client.endpoint("spam")
    assert res == "egg"


def test_api_typeerror(swh_rpc_client):
    with pytest.raises(RemoteException) as exc_info:
        swh_rpc_client.raise_typeerror()

    assert exc_info.value.args[0]["type"] == "TypeError"
    assert exc_info.value.args[0]["args"] == ["Did I pass through?"]
    assert (
        str(exc_info.value)
        == "<RemoteException 500 TypeError: ['Did I pass through?']>"
    )



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import pkg_resources
import textwrap
from unittest.mock import patch

import click
from click.testing import CliRunner
import pytest


help_msg = """Usage: swh [OPTIONS] COMMAND [ARGS]...

  Command line interface for Software Heritage.

Options:
  -l, --log-level [NOTSET|DEBUG|INFO|WARNING|ERROR|CRITICAL]
                                  Log level (defaults to INFO).
  --log-config FILENAME           Python yaml logging configuration file.
  --sentry-dsn TEXT               DSN of the Sentry instance to report to
  -h, --help                      Show this message and exit.

Notes:
  If both options are present, --log-level will override the root logger
  configuration set in --log-config.

  The --log-config YAML must conform to the logging.config.dictConfig schema
  documented at https://docs.python.org/3/library/logging.config.html.
"""


def test_swh_help(swhmain):
    runner = CliRunner()
    result = runner.invoke(swhmain, ["-h"])
    assert result.exit_code == 0
    assert result.output.startswith(help_msg)

    result = runner.invoke(swhmain, ["--help"])
    assert result.exit_code == 0
    assert result.output.startswith(help_msg)


def test_command(swhmain):
    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        click.echo("Hello SWH!")

    runner = CliRunner()
    with patch("sentry_sdk.init") as sentry_sdk_init:
        result = runner.invoke(swhmain, ["test"])
    sentry_sdk_init.assert_not_called()
    assert result.exit_code == 0
    assert result.output.strip() == "Hello SWH!"


def test_loglevel_default(caplog, swhmain):
    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        assert logging.root.level == 20
        click.echo("Hello SWH!")

    runner = CliRunner()
    result = runner.invoke(swhmain, ["test"])
    assert result.exit_code == 0
    print(result.output)
    assert result.output.strip() == """Hello SWH!"""


def test_loglevel_error(caplog, swhmain):
    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        assert logging.root.level == 40
        click.echo("Hello SWH!")

    runner = CliRunner()
    result = runner.invoke(swhmain, ["-l", "ERROR", "test"])
    assert result.exit_code == 0
    assert result.output.strip() == """Hello SWH!"""


def test_loglevel_debug(caplog, swhmain):
    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        assert logging.root.level == 10
        click.echo("Hello SWH!")

    runner = CliRunner()
    result = runner.invoke(swhmain, ["-l", "DEBUG", "test"])
    assert result.exit_code == 0
    assert result.output.strip() == """Hello SWH!"""


def test_sentry(swhmain):
    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        click.echo("Hello SWH!")

    runner = CliRunner()
    with patch("sentry_sdk.init") as sentry_sdk_init:
        result = runner.invoke(swhmain, ["--sentry-dsn", "test_dsn", "test"])
    assert result.exit_code == 0
    assert result.output.strip() == """Hello SWH!"""
    sentry_sdk_init.assert_called_once_with(
        dsn="test_dsn", debug=False, integrations=[], release=None, environment=None,
    )


def test_sentry_debug(swhmain):
    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        click.echo("Hello SWH!")

    runner = CliRunner()
    with patch("sentry_sdk.init") as sentry_sdk_init:
        result = runner.invoke(
            swhmain, ["--sentry-dsn", "test_dsn", "--sentry-debug", "test"]
        )
    assert result.exit_code == 0
    assert result.output.strip() == """Hello SWH!"""
    sentry_sdk_init.assert_called_once_with(
        dsn="test_dsn", debug=True, integrations=[], release=None, environment=None,
    )


def test_sentry_env(swhmain):
    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        click.echo("Hello SWH!")

    runner = CliRunner()
    with patch("sentry_sdk.init") as sentry_sdk_init:
        env = {
            "SWH_SENTRY_DSN": "test_dsn",
            "SWH_SENTRY_DEBUG": "1",
        }
        result = runner.invoke(swhmain, ["test"], env=env, auto_envvar_prefix="SWH")
    assert result.exit_code == 0
    assert result.output.strip() == """Hello SWH!"""
    sentry_sdk_init.assert_called_once_with(
        dsn="test_dsn", debug=True, integrations=[], release=None, environment=None,
    )


def test_sentry_env_main_package(swhmain):
    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        click.echo("Hello SWH!")

    runner = CliRunner()
    with patch("sentry_sdk.init") as sentry_sdk_init:
        env = {
            "SWH_SENTRY_DSN": "test_dsn",
            "SWH_MAIN_PACKAGE": "swh.core",
            "SWH_SENTRY_ENVIRONMENT": "tests",
        }
        result = runner.invoke(swhmain, ["test"], env=env, auto_envvar_prefix="SWH")
    assert result.exit_code == 0

    version = pkg_resources.get_distribution("swh.core").version

    assert result.output.strip() == """Hello SWH!"""
    sentry_sdk_init.assert_called_once_with(
        dsn="test_dsn",
        debug=False,
        integrations=[],
        release="swh.core@" + version,
        environment="tests",
    )


@pytest.fixture
def log_config_path(tmp_path):
    log_config = textwrap.dedent(
        """\
    ---
    version: 1
    formatters:
      formatter:
        format: 'custom format:%(name)s:%(levelname)s:%(message)s'
    handlers:
      console:
        class: logging.StreamHandler
        stream: ext://sys.stdout
        formatter: formatter
        level: DEBUG
    root:
      level: DEBUG
      handlers:
        - console
    loggers:
      dontshowdebug:
        level: INFO
    """
    )

    (tmp_path / "log_config.yml").write_text(log_config)

    yield str(tmp_path / "log_config.yml")


def test_log_config(caplog, log_config_path, swhmain):
    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        logging.debug("Root log debug")
        logging.info("Root log info")
        logging.getLogger("dontshowdebug").debug("Not shown")
        logging.getLogger("dontshowdebug").info("Shown")

    runner = CliRunner()
    result = runner.invoke(swhmain, ["--log-config", log_config_path, "test",],)

    assert result.exit_code == 0
    assert result.output.strip() == "\n".join(
        [
            "custom format:root:DEBUG:Root log debug",
            "custom format:root:INFO:Root log info",
            "custom format:dontshowdebug:INFO:Shown",
        ]
    )


def test_log_config_log_level_interaction(caplog, log_config_path, swhmain):
    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        logging.debug("Root log debug")
        logging.info("Root log info")
        logging.getLogger("dontshowdebug").debug("Not shown")
        logging.getLogger("dontshowdebug").info("Shown")

    runner = CliRunner()
    result = runner.invoke(
        swhmain, ["--log-config", log_config_path, "--log-level", "INFO", "test",],
    )

    assert result.exit_code == 0
    assert result.output.strip() == "\n".join(
        [
            "custom format:root:INFO:Root log info",
            "custom format:dontshowdebug:INFO:Shown",
        ]
    )


def test_aliased_command(swhmain):
    @swhmain.command(name="canonical-test")
    @click.pass_context
    def swhtest(ctx):
        "A test command."
        click.echo("Hello SWH!")

    swhmain.add_alias(swhtest, "othername")

    runner = CliRunner()

    # check we have only 'canonical-test' listed in the usage help msg
    result = runner.invoke(swhmain, ["-h"])
    assert result.exit_code == 0
    assert "canonical-test  A test command." in result.output
    assert "othername" not in result.output

    # check we can execute the cmd with 'canonical-test'
    result = runner.invoke(swhmain, ["canonical-test"])
    assert result.exit_code == 0
    assert result.output.strip() == """Hello SWH!"""

    # check we can also execute the cmd with the alias 'othername'
    result = runner.invoke(swhmain, ["othername"])
    assert result.exit_code == 0
    assert result.output.strip() == """Hello SWH!"""



import pytest
from hypothesis import settings

from swh.core.cli import swh as _swhmain

# define tests profile. Full documentation is at:
# https://hypothesis.readthedocs.io/en/latest/settings.html#settings-profiles
settings.register_profile("fast", max_examples=5, deadline=5000)
settings.register_profile("slow", max_examples=20, deadline=5000)


@pytest.fixture
def swhmain():
    """Yield an instance of the main `swh` click command that cleans the added
    subcommands up on teardown."""
    commands = _swhmain.commands.copy()
    aliases = _swhmain.aliases.copy()
    yield _swhmain
    _swhmain.commands = commands
    _swhmain.aliases = aliases



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This code has been imported from psycopg2, version 2.7.4,
# https://github.com/psycopg/psycopg2/tree/5afb2ce803debea9533e293eef73c92ffce95bcd
# and modified by Software Heritage.
#
# Original file: lib/extras.py
#
# psycopg2 is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.


import re
import functools

import psycopg2.extensions


def stored_procedure(stored_proc):
    """decorator to execute remote stored procedure, specified as argument

    Generally, the body of the decorated function should be empty. If it is
    not, the stored procedure will be executed first; the function body then.

    """

    def wrap(meth):
        @functools.wraps(meth)
        def _meth(self, *args, **kwargs):
            cur = kwargs.get("cur", None)
            self._cursor(cur).execute("SELECT %s()" % stored_proc)
            meth(self, *args, **kwargs)

        return _meth

    return wrap


def jsonize(value):
    """Convert a value to a psycopg2 JSON object if necessary"""
    if isinstance(value, dict):
        return psycopg2.extras.Json(value)

    return value


def _paginate(seq, page_size):
    """Consume an iterable and return it in chunks.
    Every chunk is at most `page_size`. Never return an empty chunk.
    """
    page = []
    it = iter(seq)
    while 1:
        try:
            for i in range(page_size):
                page.append(next(it))
            yield page
            page = []
        except StopIteration:
            if page:
                yield page
            return


def _split_sql(sql):
    """Split *sql* on a single ``%s`` placeholder.
    Split on the %s, perform %% replacement and return pre, post lists of
    snippets.
    """
    curr = pre = []
    post = []
    tokens = re.split(br"(%.)", sql)
    for token in tokens:
        if len(token) != 2 or token[:1] != b"%":
            curr.append(token)
            continue

        if token[1:] == b"s":
            if curr is pre:
                curr = post
            else:
                raise ValueError("the query contains more than one '%s' placeholder")
        elif token[1:] == b"%":
            curr.append(b"%")
        else:
            raise ValueError(
                "unsupported format character: '%s'"
                % token[1:].decode("ascii", "replace")
            )

    if curr is pre:
        raise ValueError("the query doesn't contain any '%s' placeholder")

    return pre, post


def execute_values_generator(cur, sql, argslist, template=None, page_size=100):
    """Execute a statement using SQL ``VALUES`` with a sequence of parameters.
    Rows returned by the query are returned through a generator.
    You need to consume the generator for the queries to be executed!

    :param cur: the cursor to use to execute the query.
    :param sql: the query to execute. It must contain a single ``%s``
        placeholder, which will be replaced by a `VALUES list`__.
        Example: ``"INSERT INTO mytable (id, f1, f2) VALUES %s"``.
    :param argslist: sequence of sequences or dictionaries with the arguments
        to send to the query. The type and content must be consistent with
        *template*.
    :param template: the snippet to merge to every item in *argslist* to
        compose the query.

        - If the *argslist* items are sequences it should contain positional
          placeholders (e.g. ``"(%s, %s, %s)"``, or ``"(%s, %s, 42)``" if there
          are constants value...).
        - If the *argslist* items are mappings it should contain named
          placeholders (e.g. ``"(%(id)s, %(f1)s, 42)"``).

        If not specified, assume the arguments are sequence and use a simple
        positional template (i.e.  ``(%s, %s, ...)``), with the number of
        placeholders sniffed by the first element in *argslist*.
    :param page_size: maximum number of *argslist* items to include in every
        statement. If there are more items the function will execute more than
        one statement.
    :param yield_from_cur: Whether to yield results from the cursor in this
        function directly.

    .. __: https://www.postgresql.org/docs/current/static/queries-values.html

    After the execution of the function the `cursor.rowcount` property will
    **not** contain a total result.
    """
    # we can't just use sql % vals because vals is bytes: if sql is bytes
    # there will be some decoding error because of stupid codec used, and Py3
    # doesn't implement % on bytes.
    if not isinstance(sql, bytes):
        sql = sql.encode(psycopg2.extensions.encodings[cur.connection.encoding])
    pre, post = _split_sql(sql)

    for page in _paginate(argslist, page_size=page_size):
        if template is None:
            template = b"(" + b",".join([b"%s"] * len(page[0])) + b")"
        parts = pre[:]
        for args in page:
            parts.append(cur.mogrify(template, args))
            parts.append(b",")
        parts[-1:] = post
        cur.execute(b"".join(parts))
        yield from cur



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import json
from typing import Any, Callable, List, Tuple
import unittest
from uuid import UUID

import arrow
import requests
import requests_mock

from swh.core.api.serializers import (
    SWHJSONDecoder,
    SWHJSONEncoder,
    msgpack_dumps,
    msgpack_loads,
    decode_response,
)


class ExtraType:
    def __init__(self, arg1, arg2):
        self.arg1 = arg1
        self.arg2 = arg2

    def __repr__(self):
        return f"ExtraType({self.arg1}, {self.arg2})"

    def __eq__(self, other):
        return isinstance(other, ExtraType) and (self.arg1, self.arg2) == (
            other.arg1,
            other.arg2,
        )


extra_encoders: List[Tuple[type, str, Callable[..., Any]]] = [
    (ExtraType, "extratype", lambda o: (o.arg1, o.arg2))
]


extra_decoders = {
    "extratype": lambda o: ExtraType(*o),
}


class Serializers(unittest.TestCase):
    def setUp(self):
        self.tz = datetime.timezone(datetime.timedelta(minutes=118))

        self.data = {
            "bytes": b"123456789\x99\xaf\xff\x00\x12",
            "datetime_tz": datetime.datetime(
                2015, 3, 4, 18, 25, 13, 1234, tzinfo=self.tz
            ),
            "datetime_utc": datetime.datetime(
                2015, 3, 4, 18, 25, 13, 1234, tzinfo=datetime.timezone.utc
            ),
            "datetime_delta": datetime.timedelta(64),
            "arrow_date": arrow.get("2018-04-25T16:17:53.533672+00:00"),
            "swhtype": "fake",
            "swh_dict": {"swhtype": 42, "d": "test"},
            "random_dict": {"swhtype": 43},
            "uuid": UUID("cdd8f804-9db6-40c3-93ab-5955d3836234"),
        }

        self.encoded_data = {
            "bytes": {"swhtype": "bytes", "d": "F)}kWH8wXmIhn8j01^"},
            "datetime_tz": {
                "swhtype": "datetime",
                "d": "2015-03-04T18:25:13.001234+01:58",
            },
            "datetime_utc": {
                "swhtype": "datetime",
                "d": "2015-03-04T18:25:13.001234+00:00",
            },
            "datetime_delta": {
                "swhtype": "timedelta",
                "d": {"days": 64, "seconds": 0, "microseconds": 0},
            },
            "arrow_date": {"swhtype": "arrow", "d": "2018-04-25T16:17:53.533672+00:00"},
            "swhtype": "fake",
            "swh_dict": {"swhtype": 42, "d": "test"},
            "random_dict": {"swhtype": 43},
            "uuid": {"swhtype": "uuid", "d": "cdd8f804-9db6-40c3-93ab-5955d3836234"},
        }

        self.legacy_msgpack = {
            "bytes": b"\xc4\x0e123456789\x99\xaf\xff\x00\x12",
            "datetime_tz": (
                b"\x82\xc4\x0c__datetime__\xc3\xc4\x01s\xd9 "
                b"2015-03-04T18:25:13.001234+01:58"
            ),
            "datetime_utc": (
                b"\x82\xc4\x0c__datetime__\xc3\xc4\x01s\xd9 "
                b"2015-03-04T18:25:13.001234+00:00"
            ),
            "datetime_delta": (
                b"\x82\xc4\r__timedelta__\xc3\xc4\x01s\x83\xa4"
                b"days@\xa7seconds\x00\xacmicroseconds\x00"
            ),
            "arrow_date": (
                b"\x82\xc4\t__arrow__\xc3\xc4\x01s\xd9 "
                b"2018-04-25T16:17:53.533672+00:00"
            ),
            "swhtype": b"\xa4fake",
            "swh_dict": b"\x82\xa7swhtype*\xa1d\xa4test",
            "random_dict": b"\x81\xa7swhtype+",
            "uuid": (
                b"\x82\xc4\x08__uuid__\xc3\xc4\x01s\xd9$"
                b"cdd8f804-9db6-40c3-93ab-5955d3836234"
            ),
        }

        self.generator = (i for i in range(5))
        self.gen_lst = list(range(5))

    def test_round_trip_json(self):
        data = json.dumps(self.data, cls=SWHJSONEncoder)
        self.assertEqual(self.data, json.loads(data, cls=SWHJSONDecoder))

    def test_round_trip_json_extra_types(self):
        original_data = [ExtraType("baz", self.data), "qux"]

        data = json.dumps(
            original_data, cls=SWHJSONEncoder, extra_encoders=extra_encoders
        )
        self.assertEqual(
            original_data,
            json.loads(data, cls=SWHJSONDecoder, extra_decoders=extra_decoders),
        )

    def test_encode_swh_json(self):
        data = json.dumps(self.data, cls=SWHJSONEncoder)
        self.assertEqual(self.encoded_data, json.loads(data))

    def test_round_trip_msgpack(self):
        original_data = {
            **self.data,
            "none_dict_key": {None: 42},
            "long_int_is_loooong": 10000000000000000000000000000000,
        }
        data = msgpack_dumps(original_data)
        self.assertEqual(original_data, msgpack_loads(data))

    def test_round_trip_msgpack_extra_types(self):
        original_data = [ExtraType("baz", self.data), "qux"]

        data = msgpack_dumps(original_data, extra_encoders=extra_encoders)
        self.assertEqual(
            original_data, msgpack_loads(data, extra_decoders=extra_decoders)
        )

    def test_generator_json(self):
        data = json.dumps(self.generator, cls=SWHJSONEncoder)
        self.assertEqual(self.gen_lst, json.loads(data, cls=SWHJSONDecoder))

    def test_generator_msgpack(self):
        data = msgpack_dumps(self.generator)
        self.assertEqual(self.gen_lst, msgpack_loads(data))

    @requests_mock.Mocker()
    def test_decode_response_json(self, mock_requests):
        mock_requests.get(
            "https://example.org/test/data",
            json=self.encoded_data,
            headers={"content-type": "application/json"},
        )
        response = requests.get("https://example.org/test/data")
        assert decode_response(response) == self.data

    def test_decode_legacy_msgpack(self):
        for k, v in self.legacy_msgpack.items():
            assert msgpack_loads(v) == self.data[k]

    def test_encode_native_datetime(self):
        dt = datetime.datetime(2015, 1, 1, 12, 4, 42, 231455)
        with self.assertRaises(ValueError, matches="naive datetime"):
            msgpack_dumps(dt)

    def test_decode_naive_datetime(self):
        expected_dt = datetime.datetime(2015, 1, 1, 12, 4, 42, 231455)

        # Current encoding
        assert (
            msgpack_loads(
                b"\x82\xc4\x07swhtype\xa8datetime\xc4\x01d\xba"
                b"2015-01-01T12:04:42.231455"
            )
            == expected_dt
        )

        # Legacy encoding
        assert (
            msgpack_loads(
                b"\x82\xc4\x0c__datetime__\xc3\xc4\x01s\xba"
                b"2015-01-01T12:04:42.231455"
            )
            == expected_dt
        )



#

from click.testing import CliRunner

from swh.core.cli.db import db as swhdb


help_msg = """Usage: swh [OPTIONS] COMMAND [ARGS]...

  Command line interface for Software Heritage.

Options:
  -l, --log-level [NOTSET|DEBUG|INFO|WARNING|ERROR|CRITICAL]
                                  Log level (defaults to INFO).
  --log-config FILENAME           Python yaml logging configuration file.
  --sentry-dsn TEXT               DSN of the Sentry instance to report to
  -h, --help                      Show this message and exit.

Notes:
  If both options are present, --log-level will override the root logger
  configuration set in --log-config.

  The --log-config YAML must conform to the logging.config.dictConfig schema
  documented at https://docs.python.org/3/library/logging.config.html.

Commands:
  db  Software Heritage database generic tools.
"""


def test_swh_help(swhmain):
    swhmain.add_command(swhdb)
    runner = CliRunner()
    result = runner.invoke(swhmain, ["-h"])
    assert result.exit_code == 0
    assert result.output == help_msg


help_db_msg = """Usage: swh db [OPTIONS] COMMAND [ARGS]...

  Software Heritage database generic tools.

Options:
  -C, --config-file FILE  Configuration file.
  -h, --help              Show this message and exit.

Commands:
  init  Initialize the database for every Software Heritage module found in...
"""


def test_swh_db_help(swhmain):
    swhmain.add_command(swhdb)
    runner = CliRunner()
    result = runner.invoke(swhmain, ["db", "-h"])
    assert result.exit_code == 0
    assert result.output == help_db_msg



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from os import path


DATADIR = path.join(path.abspath(path.dirname(__file__)), "data")


@pytest.fixture
def datadir():
    return DATADIR



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import shutil
import stat
import tarfile
import zipfile

from subprocess import run

from . import utils


def _unpack_tar(tarpath: str, extract_dir: str) -> str:
    """Unpack tarballs unsupported by the standard python library. Examples
    include tar.Z, tar.lz, tar.x, etc....

    As this implementation relies on the `tar` command, this function supports
    the same compression the tar command supports.

    This expects the `extract_dir` to exist.

    Raises

        shutil.ReadError in case of issue uncompressing the archive (tarpath
        does not exist, extract_dir does not exist, etc...)

    Returns
        full path to the uncompressed directory.

    """
    try:
        run(["tar", "xf", tarpath, "-C", extract_dir], check=True)
        return extract_dir
    except Exception as e:
        raise shutil.ReadError(
            f"Unable to uncompress {tarpath} to {extract_dir}. Reason: {e}"
        )


def register_new_archive_formats():
    """Register new archive formats to uncompress

    """
    registered_formats = [f[0] for f in shutil.get_unpack_formats()]
    for name, extensions, function in ADDITIONAL_ARCHIVE_FORMATS:
        if name in registered_formats:
            continue
        shutil.register_unpack_format(name, extensions, function)


def uncompress(tarpath: str, dest: str):
    """Uncompress tarpath to dest folder if tarball is supported.

       Note that this fixes permissions after successfully
       uncompressing the archive.

    Args:
        tarpath: path to tarball to uncompress
        dest: the destination folder where to uncompress the tarball

    Returns:
        The nature of the tarball, zip or tar.

    Raises:
        ValueError when a problem occurs during unpacking

    """
    try:
        shutil.unpack_archive(tarpath, extract_dir=dest)
    except shutil.ReadError as e:
        raise ValueError(f"Problem during unpacking {tarpath}. Reason: {e}")

    # Fix permissions
    for dirpath, _, fnames in os.walk(dest):
        os.chmod(dirpath, 0o755)
        for fname in fnames:
            fpath = os.path.join(dirpath, fname)
            if not os.path.islink(fpath):
                fpath_exec = os.stat(fpath).st_mode & stat.S_IXUSR
                if not fpath_exec:
                    os.chmod(fpath, 0o644)


def _ls(rootdir):
    """Generator of filepath, filename from rootdir.

    """
    for dirpath, dirnames, fnames in os.walk(rootdir):
        for fname in dirnames + fnames:
            fpath = os.path.join(dirpath, fname)
            fname = utils.commonname(rootdir, fpath)
            yield fpath, fname


def _compress_zip(tarpath, files):
    """Compress dirpath's content as tarpath.

    """
    with zipfile.ZipFile(tarpath, "w") as z:
        for fpath, fname in files:
            z.write(fpath, arcname=fname)


def _compress_tar(tarpath, files):
    """Compress dirpath's content as tarpath.

    """
    with tarfile.open(tarpath, "w:bz2") as t:
        for fpath, fname in files:
            t.add(fpath, arcname=fname, recursive=False)


def compress(tarpath, nature, dirpath_or_files):
    """Create a tarball tarpath with nature nature.
    The content of the tarball is either dirpath's content (if representing
    a directory path) or dirpath's iterable contents.

    Compress the directory dirpath's content to a tarball.
    The tarball being dumped at tarpath.
    The nature of the tarball is determined by the nature argument.

    """
    if isinstance(dirpath_or_files, str):
        files = _ls(dirpath_or_files)
    else:  # iterable of 'filepath, filename'
        files = dirpath_or_files

    if nature == "zip":
        _compress_zip(tarpath, files)
    else:
        _compress_tar(tarpath, files)

    return tarpath


# Additional uncompression archive format support
ADDITIONAL_ARCHIVE_FORMATS = [
    # name  , extensions, function
    ("tar.Z|x", [".tar.Z", ".tar.x"], _unpack_tar),
    # FIXME: make this optional depending on the runtime lzip package install
    ("tar.lz", [".tar.lz"], _unpack_tar),
]

register_new_archive_formats()



#!/usr/bin/env python3
# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
from setuptools import setup, find_packages

from os import path
from io import open

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


def parse_requirements(*names):
    requirements = []
    for name in names:
        if name:
            reqf = "requirements-%s.txt" % name
        else:
            reqf = "requirements.txt"

        if not os.path.exists(reqf):
            return requirements

        with open(reqf) as f:
            for line in f.readlines():
                line = line.strip()
                if not line or line.startswith("#"):
                    continue
                requirements.append(line)
    return requirements


setup(
    name="swh.core",
    description="Software Heritage core utilities",
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.7",
    author="Software Heritage developers",
    author_email="swh-devel@inria.fr",
    url="https://forge.softwareheritage.org/diffusion/DCORE/",
    packages=find_packages(),
    py_modules=["pytest_swh_core"],
    scripts=[],
    install_requires=parse_requirements(None, "swh"),
    setup_requires=["vcversioner"],
    extras_require={
        "testing-core": parse_requirements("test"),
        "logging": parse_requirements("logging"),
        "db": parse_requirements("db"),
        "testing-db": parse_requirements("test-db"),
        "http": parse_requirements("http"),
        # kitchen sink, please do not use
        "testing": parse_requirements("test", "test-db", "db", "http", "logging"),
    },
    vcversioner={},
    include_package_data=True,
    entry_points="""
        [console_scripts]
        swh=swh.core.cli:main
        swh-db-init=swh.core.cli.db:db_init
        [swh.cli.subcommands]
        db=swh.core.cli.db:db
        db-init=swh.core.cli.db:db_init
        [pytest11]
        pytest_swh_core = swh.core.pytest_plugin
    """,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
    ],
    project_urls={
        "Bug Reports": "https://forge.softwareheritage.org/maniphest",
        "Funding": "https://www.softwareheritage.org/donate",
        "Source": "https://forge.softwareheritage.org/source/swh-core",
        "Documentation": "https://docs.softwareheritage.org/devel/swh-core/",
    },
)



# Copyright (C) 2017-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import OrderedDict
import logging
from typing import Tuple, Type

import aiohttp.web
from deprecated import deprecated
import multidict

from .serializers import msgpack_dumps, msgpack_loads
from .serializers import json_dumps, json_loads
from .serializers import exception_to_dict

from aiohttp_utils import negotiation, Response


def encode_msgpack(data, **kwargs):
    return aiohttp.web.Response(
        body=msgpack_dumps(data),
        headers=multidict.MultiDict({"Content-Type": "application/x-msgpack"}),
        **kwargs,
    )


encode_data_server = Response


def render_msgpack(request, data):
    return msgpack_dumps(data)


def render_json(request, data):
    return json_dumps(data)


async def decode_request(request):
    content_type = request.headers.get("Content-Type").split(";")[0].strip()
    data = await request.read()
    if not data:
        return {}
    if content_type == "application/x-msgpack":
        r = msgpack_loads(data)
    elif content_type == "application/json":
        r = json_loads(data)
    else:
        raise ValueError("Wrong content type `%s` for API request" % content_type)
    return r


async def error_middleware(app, handler):
    async def middleware_handler(request):
        try:
            return await handler(request)
        except Exception as e:
            if isinstance(e, aiohttp.web.HTTPException):
                raise
            logging.exception(e)
            res = exception_to_dict(e)
            if isinstance(e, app.client_exception_classes):
                status = 400
            else:
                status = 500
            return encode_data_server(res, status=status)

    return middleware_handler


class RPCServerApp(aiohttp.web.Application):
    client_exception_classes: Tuple[Type[Exception], ...] = ()
    """Exceptions that should be handled as a client error (eg. object not
    found, invalid argument)"""

    def __init__(self, *args, middlewares=(), **kwargs):
        middlewares = (error_middleware,) + middlewares
        # renderers are sorted in order of increasing desirability (!)
        # see mimeparse.best_match() docstring.
        renderers = OrderedDict(
            [
                ("application/json", render_json),
                ("application/x-msgpack", render_msgpack),
            ]
        )
        nego_middleware = negotiation.negotiation_middleware(
            renderers=renderers, force_rendering=True
        )
        middlewares = (nego_middleware,) + middlewares

        super().__init__(*args, middlewares=middlewares, **kwargs)


@deprecated(version="0.0.64", reason="Use the RPCServerApp instead")
class SWHRemoteAPI(RPCServerApp):
    pass



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import glob
import subprocess

import psycopg2

from typing import Dict, Iterable, Optional, Tuple, Union

from swh.core.utils import numfile_sortkey as sortkey


DB_DUMP_TYPES = {".sql": "psql", ".dump": "pg_dump"}  # type: Dict[str, str]


def swh_db_version(dbname_or_service):
    """Retrieve the swh version if any. In case of the db not initialized,
    this returns None. Otherwise, this returns the db's version.

    Args:
        dbname_or_service (str): The db's name or service

    Returns:
        Optional[Int]: Either the db's version or None

    """
    query = "select version from dbversion order by dbversion desc limit 1"
    cmd = [
        "psql",
        "--tuples-only",
        "--no-psqlrc",
        "--quiet",
        "-v",
        "ON_ERROR_STOP=1",
        "--command=%s" % query,
        dbname_or_service,
    ]

    try:
        r = subprocess.run(
            cmd, check=True, stdout=subprocess.PIPE, universal_newlines=True
        )
        result = int(r.stdout.strip())
    except Exception:  # db not initialized
        result = None
    return result


def pg_restore(dbname, dumpfile, dumptype="pg_dump"):
    """
    Args:
        dbname: name of the DB to restore into
        dumpfile: path of the dump file
        dumptype: one of 'pg_dump' (for binary dumps), 'psql' (for SQL dumps)
    """
    assert dumptype in ["pg_dump", "psql"]
    if dumptype == "pg_dump":
        subprocess.check_call(
            [
                "pg_restore",
                "--no-owner",
                "--no-privileges",
                "--dbname",
                dbname,
                dumpfile,
            ]
        )
    elif dumptype == "psql":
        subprocess.check_call(
            [
                "psql",
                "--quiet",
                "--no-psqlrc",
                "-v",
                "ON_ERROR_STOP=1",
                "-f",
                dumpfile,
                dbname,
            ]
        )


def pg_dump(dbname, dumpfile):
    subprocess.check_call(
        ["pg_dump", "--no-owner", "--no-privileges", "-Fc", "-f", dumpfile, dbname]
    )


def pg_dropdb(dbname):
    subprocess.check_call(["dropdb", dbname])


def pg_createdb(dbname, check=True):
    """Create a db. If check is True and the db already exists, this will
       raise an exception (original behavior). If check is False and
       the db already exists, this will fail silently. If the db does
       not exist, the db will be created.

    """
    subprocess.run(["createdb", dbname], check=check)


def db_create(dbname, dumps=None):
    """create the test DB and load the test data dumps into it

    dumps is an iterable of couples (dump_file, dump_type).

    context: setUpClass

    """
    try:
        pg_createdb(dbname)
    except subprocess.CalledProcessError:  # try recovering once, in case
        pg_dropdb(dbname)  # the db already existed
        pg_createdb(dbname)
    for dump, dtype in dumps:
        pg_restore(dbname, dump, dtype)
    return dbname


def db_destroy(dbname):
    """destroy the test DB

    context: tearDownClass

    """
    pg_dropdb(dbname)


def db_connect(dbname):
    """connect to the test DB and open a cursor

    context: setUp

    """
    conn = psycopg2.connect("dbname=" + dbname)
    return {"conn": conn, "cursor": conn.cursor()}


def db_close(conn):
    """rollback current transaction and disconnect from the test DB

    context: tearDown

    """
    if not conn.closed:
        conn.rollback()
        conn.close()


class DbTestConn:
    def __init__(self, dbname):
        self.dbname = dbname

    def __enter__(self):
        self.db_setup = db_connect(self.dbname)
        self.conn = self.db_setup["conn"]
        self.cursor = self.db_setup["cursor"]
        return self

    def __exit__(self, *_):
        db_close(self.conn)


class DbTestContext:
    def __init__(self, name="softwareheritage-test", dumps=None):
        self.dbname = name
        self.dumps = dumps

    def __enter__(self):
        db_create(dbname=self.dbname, dumps=self.dumps)
        return self

    def __exit__(self, *_):
        db_destroy(self.dbname)


class DbTestFixture:
    """Mix this in a test subject class to get DB testing support.

    Use the class method add_db() to add a new database to be tested.
    Using this will create a DbTestConn entry in the `test_db` dictionary for
    all the tests, indexed by the name of the database.

    Example:

    class TestDb(DbTestFixture, unittest.TestCase):
        @classmethod
        def setUpClass(cls):
            cls.add_db('db_name', DUMP)
            super().setUpClass()

        def setUp(self):
            db = self.test_db['db_name']
            print('conn: {}, cursor: {}'.format(db.conn, db.cursor))

    To ensure test isolation, each test method of the test case class will
    execute in its own connection, cursor, and transaction.

    Note that if you want to define setup/teardown methods, you need to
    explicitly call super() to ensure that the fixture setup/teardown methods
    are invoked. Here is an example where all setup/teardown methods are
    defined in a test case:

        class TestDb(DbTestFixture, unittest.TestCase):
            @classmethod
            def setUpClass(cls):
                # your add_db() calls here
                super().setUpClass()
                # your class setup code here

            def setUp(self):
                super().setUp()
                # your instance setup code here

            def tearDown(self):
                # your instance teardown code here
                super().tearDown()

            @classmethod
            def tearDownClass(cls):
                # your class teardown code here
                super().tearDownClass()

    """

    _DB_DUMP_LIST = {}  # type: Dict[str, Iterable[Tuple[str, str]]]
    _DB_LIST = {}  # type: Dict[str, DbTestContext]
    DB_TEST_FIXTURE_IMPORTED = True

    @classmethod
    def add_db(cls, name="softwareheritage-test", dumps=None):
        cls._DB_DUMP_LIST[name] = dumps

    @classmethod
    def setUpClass(cls):
        for name, dumps in cls._DB_DUMP_LIST.items():
            cls._DB_LIST[name] = DbTestContext(name, dumps)
            cls._DB_LIST[name].__enter__()
        super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        for name, context in cls._DB_LIST.items():
            context.__exit__()

    def setUp(self, *args, **kwargs):
        self.test_db = {}
        for name in self._DB_LIST.keys():
            self.test_db[name] = DbTestConn(name)
            self.test_db[name].__enter__()
        super().setUp(*args, **kwargs)

    def tearDown(self):
        super().tearDown()
        for name in self._DB_LIST.keys():
            self.test_db[name].__exit__()

    def reset_db_tables(self, name, excluded=None):
        db = self.test_db[name]
        conn = db.conn
        cursor = db.cursor

        cursor.execute(
            """SELECT table_name FROM information_schema.tables
                          WHERE table_schema = %s""",
            ("public",),
        )

        tables = set(table for (table,) in cursor.fetchall())
        if excluded is not None:
            tables -= set(excluded)

        for table in tables:
            cursor.execute("truncate table %s cascade" % table)

        conn.commit()


class SingleDbTestFixture(DbTestFixture):
    """Simplified fixture like DbTest but that can only handle a single DB.

    Gives access to shortcuts like self.cursor and self.conn.

    DO NOT use this with other fixtures that need to access databases, like
    StorageTestFixture.

    The class can override the following class attributes:
        TEST_DB_NAME: name of the DB used for testing
        TEST_DB_DUMP: DB dump to be restored before running test methods; can
            be set to None if no restore from dump is required.
            If the dump file name endswith"
            - '.sql' it will be loaded via psql,
            - '.dump' it will be loaded via pg_restore.
            Other file extensions will be ignored.
            Can be a string or a list of strings; each path will be expanded
            using glob pattern matching.

    The test case class will then have the following attributes, accessible via
    self:

        dbname: name of the test database
        conn: psycopg2 connection object
        cursor: open psycopg2 cursor to the DB
    """

    TEST_DB_NAME = "softwareheritage-test"
    TEST_DB_DUMP = None  # type: Optional[Union[str, Iterable[str]]]

    @classmethod
    def setUpClass(cls):
        cls.dbname = cls.TEST_DB_NAME  # XXX to kill?

        dump_files = cls.TEST_DB_DUMP
        if dump_files is None:
            dump_files = []
        elif isinstance(dump_files, str):
            dump_files = [dump_files]
        all_dump_files = []
        for files in dump_files:
            all_dump_files.extend(sorted(glob.glob(files), key=sortkey))

        all_dump_files = [
            (x, DB_DUMP_TYPES[os.path.splitext(x)[1]]) for x in all_dump_files
        ]

        cls.add_db(name=cls.TEST_DB_NAME, dumps=all_dump_files)
        super().setUpClass()

    def setUp(self, *args, **kwargs):
        super().setUp(*args, **kwargs)

        db = self.test_db[self.TEST_DB_NAME]
        self.conn = db.conn
        self.cursor = db.cursor



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import logging.config
import signal

import click
import pkg_resources
import yaml

from ..sentry import init_sentry

LOG_LEVEL_NAMES = ["NOTSET", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])

logger = logging.getLogger(__name__)


class AliasedGroup(click.Group):
    """A simple Group that supports command aliases, as well as notes related to
    options"""

    def __init__(self, name=None, commands=None, **attrs):
        self.option_notes = attrs.pop("option_notes", None)
        self.aliases = {}
        super().__init__(name, commands, **attrs)

    def get_command(self, ctx, cmd_name):
        return super().get_command(ctx, self.aliases.get(cmd_name, cmd_name))

    def add_alias(self, name, alias):
        if not isinstance(name, str):
            name = name.name
        self.aliases[alias] = name

    def format_options(self, ctx, formatter):
        click.Command.format_options(self, ctx, formatter)
        if self.option_notes:
            with formatter.section("Notes"):
                formatter.write_text(self.option_notes)
        self.format_commands(ctx, formatter)


def clean_exit_on_signal(signal, frame):
    """Raise a SystemExit exception to let command-line clients wind themselves
    down on exit"""
    raise SystemExit(0)


@click.group(
    context_settings=CONTEXT_SETTINGS,
    cls=AliasedGroup,
    option_notes="""\
If both options are present, --log-level will override the root logger
configuration set in --log-config.

The --log-config YAML must conform to the logging.config.dictConfig schema
documented at https://docs.python.org/3/library/logging.config.html.
""",
)
@click.option(
    "--log-level",
    "-l",
    default=None,
    type=click.Choice(LOG_LEVEL_NAMES),
    help="Log level (defaults to INFO).",
)
@click.option(
    "--log-config",
    default=None,
    type=click.File("r"),
    help="Python yaml logging configuration file.",
)
@click.option(
    "--sentry-dsn", default=None, help="DSN of the Sentry instance to report to"
)
@click.option(
    "--sentry-debug/--no-sentry-debug",
    default=False,
    hidden=True,
    help="Enable debugging of sentry",
)
@click.pass_context
def swh(ctx, log_level, log_config, sentry_dsn, sentry_debug):
    """Command line interface for Software Heritage.
    """
    signal.signal(signal.SIGTERM, clean_exit_on_signal)
    signal.signal(signal.SIGINT, clean_exit_on_signal)

    init_sentry(sentry_dsn, debug=sentry_debug)

    if log_level is None and log_config is None:
        log_level = "INFO"

    if log_config:
        logging.config.dictConfig(yaml.safe_load(log_config.read()))

    if log_level:
        log_level = logging.getLevelName(log_level)
        logging.root.setLevel(log_level)

    ctx.ensure_object(dict)
    ctx.obj["log_level"] = log_level


def main():
    # Even though swh() sets up logging, we need an earlier basic logging setup
    # for the next few logging statements
    logging.basicConfig()
    # load plugins that define cli sub commands
    for entry_point in pkg_resources.iter_entry_points("swh.cli.subcommands"):
        try:
            cmd = entry_point.load()
            swh.add_command(cmd, name=entry_point.name)
        except Exception as e:
            logger.warning("Could not load subcommand %s: %s", entry_point.name, str(e))

    return swh(auto_envvar_prefix="SWH")


if __name__ == "__main__":
    main()
